import os
import openai
import logging
from openai.error import Timeout, APIError, APIConnectionError, RateLimitError, ServiceUnavailableError
from tenacity import before_sleep_log, retry as tenacity_retry, retry_if_exception_type as retry_if, stop_after_attempt, wait_exponential
from typing import Callable, ContextManager
from enum import Enum

from rich import print as rprint
from rich.spinner import Spinner
from rich.live import Live


logger = logging.getLogger(__name__)
retry = tenacity_retry(
    reraise=True,
    stop=stop_after_attempt(4),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    retry=retry_if((Timeout, APIError, APIConnectionError, RateLimitError, ServiceUnavailableError)),
    before_sleep=before_sleep_log(logger, logging.WARNING),
)

class Role(str, Enum):
    system = 'system'
    assistant = 'assistant'
    user = 'user'

class Message(dict):
    """Message format for communicating with the OpenAI API."""
    def __init__(self, role:Role, content:str):
        super().__init__(role=role.value, content=content)

class ContextMessage(Message):
    """Simple wrapper around a message that adds an id and optional lifetime."""
    def __init__(self, role:Role, content:str, id:int, lifetime:int|None=None):
        super().__init__(role=role, content=content)
        self.id = id
        self.lifetime = lifetime


def cli_spinner(): 
    return Live(Spinner('dots', speed=2, text="thinking..."), refresh_per_second=30, transient=True)
class no_spinner:
    def __enter__(self): pass
    def __exit__(self, *args): pass

class Agent:
    def __init__(self, *, model:str='gpt-4', prompt:str="You are a helpful assistant.", api_key:str|None=None, spinner:Callable[[], ContextManager]|None=cli_spinner, rich_print:bool=True):
        """
        Agent class for managing communication with OpenAI's API.

        Args:
            model (str, optional): The name of the model to use. Defaults to 'gpt-4'. At present, GPT-4 is the only model that works reliably.
            prompt (str, optional): The prompt to use when starting a new conversation. Defaults to "You are a helpful assistant.".
            api_key (str, optional): The OpenAI API key to use. Defaults to None. If None, the API key will be read from the OPENAI_API_KEY environment variable.
            spinner ((fn -> ContextManager) | None, optional): A function that returns a context manager that is run every time the LLM is generating a response. Defaults to cli_spinner which is used to display a spinner in the terminal.
            rich_print (bool, optional): Whether to use rich to print messages. Defaults to True. Can also be set via the DISABLE_RICH_PRINT environment variable.

        Raises:
            Exception: If no API key is given.
        """

        self.rich_print = bool(rich_print and not os.environ.get("DISABLE_RICH_PRINT", False))
        self.model = model
        self.system_message = Message(role=Role.system, content=prompt)
        self.messages: list[Message] = []
        if spinner is not None and self.rich_print:
            self.spinner = spinner
        else:
            self.spinner = no_spinner

        # use to generate unique ids for context messages
        self._current_context_id = 0

        # check that an api key was given, and set it
        if api_key is None:
            api_key = os.environ.get('OPENAI_API_KEY', None)
        if not api_key:
            raise Exception("No OpenAI API key given. Please set the OPENAI_API_KEY environment variable or pass the api_key argument to the Agent constructor.")
        openai.api_key = api_key

    def print(self, *args, **kwargs):
        if self.rich_print:
            rprint(*args, **kwargs)
        else:
            print(*args, **kwargs)

    def new_context_id(self) -> int:
        """Generate a new context id."""
        self._current_context_id += 1
        return self._current_context_id

    def add_context(self, context:str, *, lifetime:int|None=None) -> int:
        """
        Inject a context message to the agent's conversation.

        Useful for providing the agent with information relevant to the current conversation, e.g. tool state, environment info, etc.
        If a lifetime is specified, the context message will automatically be deleted from the chat history after that many steps.
        A context message can be deleted manually by calling clear_context() with the id of the context message.

        Args:
            context (str): The context to add to the agent's conversation.
            lifetime (int, optional): The number of time steps the context will live for. Defaults to None (i.e. it will never be removed).

        Returns:
            int: The id of the context message.
        """
        context_message = ContextMessage(role=Role.system, content=context, id=self.new_context_id(), lifetime=lifetime)
        self.messages.append(context_message)
        return context_message.id

    def update_timed_context(self) -> None:
        """
        Update the lifetimes of all timed contexts, and remove any that have expired.
        This should be called after every LLM response.
        """
        #decrement lifetimes of all timed context messages
        for message in self.messages:
            if isinstance(message, ContextMessage) and message.lifetime is not None:
                message.lifetime -= 1

        #remove expired context messages
        new_messages = []
        for message in self.messages:
            if isinstance(message, ContextMessage) and message.lifetime == 0:
                continue
            new_messages.append(message)
        self.messages = new_messages


    def clear_context(self, id:int) -> None:
        """
        Remove a single context message from the agent's conversation.

        Args:
            id (int): The id of the context message to remove.
        """
        new_messages = []
        for message in self.messages:
            if isinstance(message, ContextMessage) and message.id == id:
                continue
            new_messages.append(message)
        self.messages = new_messages


    def clear_all_context(self) -> None:
        """Remove all context messages from the agent's conversation."""
        self.messages = [message for message in self.messages if not isinstance(message, ContextMessage)]
    
    def query(self, message:str) -> str:
        """Send a user query to the agent. Returns the agent's response"""
        self.messages.append(Message(role=Role.user, content=message))
        return self.execute()
    
    def observe(self, observation:str) -> str:
        """Send a system/tool observation to the agent. Returns the agent's response"""
        self.messages.append(Message(role=Role.system, content=observation))
        return self.execute()
    
    def error(self, error:str, drop_error:bool=True) -> str:
        """
        Send an error message to the agent. Returns the agent's response.
        
        Args:
            error (str): The error message to send to the agent.
            drop_error (bool, optional): If True, the error message and LLMs bad input will be dropped from the chat history. Defaults to `True`.
        """
        self.messages.append(Message(role=Role.system, content=f'ERROR: {error}'))
        result = self.execute()

        # Drop error + LLM's bad input from chat history
        if drop_error:
            del self.messages[-3:-1]

        return result
    
    @retry
    def execute(self) -> str:
        with self.spinner():
            completion = openai.ChatCompletion.create(
                model=self.model, 
                messages=[self.system_message] + self.messages,
                temperature=0,
            )
        
        # grab the response and add it to the chat history
        result = completion.choices[0].message.content
        self.messages.append(Message(role=Role.assistant, content=result))

        # remove any timed contexts that have expired
        self.update_timed_context()

        return result

    @retry
    def oneshot(self, prompt:str, query:str) -> str:
        """
        Send a user query to the agent. Returns the agent's response.
        This method ignores any previous conversation history, as well as the existing prompt.
        The output is the raw LLM text withouth any postprocessing, so you'll need to handle parsing it yourself.

        Args:
            prompt (str): The prompt to use when starting a new conversation.
            query (str): The user query to send to the agent.

        Returns:
            str: The agent's response to the user query.
        """
        with self.spinner():
            completion = openai.ChatCompletion.create(
                model=self.model, 
                messages=[Message(role=Role.system, content=prompt), Message(role=Role.user, content=query)],
                temperature=0,
            )

        # return the agent's response
        result = completion.choices[0].message.content
        return result



import sys
from io import StringIO
from types import ModuleType
from typing import Any
from archytas.utils import get_local_name

import pdb



class Python:
    def __init__(self):
        self.locals: dict[str, Any] = {}
        self._added_locals: dict[str, Any] = {} # keep track of any modules/classes/functions/etc. added to locals
        self.imports: list[str] = []            # keep track of any imports added
        self.all_scripts: list[str] = []        # keep track of all scripts run

    def update_locals(self, new_locals:dict):
        self.locals.update(new_locals)

    def add_locals(self, new_locals:list[type|ModuleType|Any], outer_locals:dict[str, Any]):
        for local in new_locals:
            if isinstance(local, type):
                update = {local.__name__: local}
            elif isinstance(local, ModuleType):
                name = get_local_name(local, outer_locals)
                update = {name: local}
            else:
                #TODO: not sure if best way to handle other types...
                name = get_local_name(local, outer_locals)
                update = {name: local}

            self._added_locals.update(update)
            self.locals.update(update)


    def run_script(self, script:str):
        #capture any stdout/stderr from the script
        captured_stdout = StringIO()
        captured_stderr = StringIO()
        sys.stdout = captured_stdout
        sys.stderr = captured_stderr

        #save the script text
        self.all_scripts.append(script)

        #run the script
        try:
            exec(script, self.locals)
        except Exception as e:
            sys.stderr.write(str(e))

        #restore stdout/stderr
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        return captured_stdout.getvalue(), captured_stderr.getvalue()


    # def add_imports(self, imports:list[str]):
    #     #TODO: this should import them into the locals dict by calling exec() on the import statements
    #     #      also handle splitting up multiple imports in a single string
    #     #TODO: this could also be a more generic add_code() method
    #     self.imports.extend(imports)
        


def main():
    import numpy as np

    class MyClass:
        def __init__(self, value):
            self.value = value


    p = Python()

    #prog0
    out,err = p.run_script('''
print('Hello world!')
    ''')
    print(f'out: {out}')
    print(f'err: {err}')


    #prog1
    out, err = p.run_script('''
raise Exception('Hello world!')
    ''')
    print(f'out: {out}')
    print(f'err: {err}')

    #prog2
    p.add_locals([np], locals())
    out,err = p.run_script('''
a = np.array([1, 2, 3])
b = np.array([4, 5, 6])
    ''')
    print(f'out: {out}')
    print(f'err: {err}')
    print(f'a: {p.locals["a"]}')
    print(f'b: {p.locals["b"]}')


    p.add_locals([MyClass], locals())

    out,err = p.run_script('''
result = np.dot(a, b)
my_obj = MyClass(np.array([1, 2, 3]))
    ''')
    print(f'out: {out}')
    print(f'err: {err}')
    print(f'result: {p.locals["result"]}')
    print(f'my_obj: {p.locals["my_obj"]}')



if __name__ == '__main__':
    main()


from typing import Any

def get_local_name(val:Any, locals:dict[str, Any]) -> str:
    """
    Determines the name of a local variable by finding a matching value in the given locals dict

    Performs a linear search over locals, so may not be performant if done frequently.
    
    Args:
        val (Any): The value to find the name of. This can be a module, class, function, variable, etc.
        locals (dict[str, Any]): The locals dict to search for the value in. This is usually the locals() dict from the root scope.

    Returns:
        str: The name of the local variable

    Raises:
        ValueError: If the value is not found in the locals dict
    """
    for name, obj in locals.items():
        if obj is val:
            return name
    raise ValueError(f'Value {val} not found in locals')




class InstanceMethod:
    """
    A descriptor that wraps a method so that it can be accessed from an instance of the class, but not from the class itself.
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        if instance is None:
            raise TypeError("This method should only be accessed from an instance of the class")
        return self.func.__get__(instance, owner)


from typing import Any
from archytas.tool_utils import tool, toolset, is_tool
from archytas.python import Python


@tool()
def ask_user(query:str) -> str:
    """
    Ask the user a question and get their response. 
    
    You should ask the user a question if you do not have enough information to complete the task, and there is no suitable tool to help you.
    
    Args:
        query (str): The question to ask the user

    Returns:
        str: The user's response
    """
    return input(f'{query} $ ')



from datetime import datetime
import pytz
@tool(name='datetime')
def datetime_tool(format:str='%Y-%m-%d %H:%M:%S %Z', timezone:str='UTC') -> str:
    """
    Get the current date and time. 
    
    Args:
        format (str, optional): The format to return the date and time in. Defaults to '%Y-%m-%d %H:%M:%S %Z'.
        timezone (str, optional): The timezone to return the date and time in. Defaults to 'UTC'.

    Returns:
        str: The current date and time in the specified format
    """
    # TODO: See https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes for more information.
    # TODO: list of valid timezones: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    
    tz = pytz.timezone(timezone)
    return datetime.now(tz).strftime(format)


@tool()
def timestamp() -> float:
    """
    Returns the current unix timestamp in seconds
    
    Returns:
        float: The current unix timestamp in seconds 

    Examples:
        >>> timestamp()
        1681445698.726113
    """
    return datetime.now().timestamp()


@toolset()
class PythonTool:
    """
    Tool for running python code. If the user asks you to write code, you can run it here.
    """
    def __init__(self, locals:dict[str,Any]|None=None, prelude:str=''):
        """
        Create a PythonTool instance

        Args:
            locals (dict[str,Any], optional): A dictionary of variables/classes/functions/modules/etc. to add to the python environment. Defaults to {}. @tools will be correctly extracted from their wrappers so they can be used in the environment.
            prelude (str, optional): Code to run before any other code. Defaults to ''. This could be used to import modules, or define functions/classes/etc. that will be used in the environment.        
        """
        # create the python env instance
        self.env = Python()
        
        if locals is None:
            locals = {}

        prompt_chunks = []
        
        # collect any @tools from the locals, unwrap them
        # collect the description for each tool or function
        # TODO: make class tools collect the docstring for all tool methods
        env_update = {}
        for name, obj in locals.items():
            if is_tool(obj):
                env_update[name] = obj
                prompt_chunks.append(f'{name} = {obj.__doc__}')
            else:
                env_update[name] = obj
                prompt_chunks.append(f'{name} = {obj} ({type(obj)})')
        
        # update the env with the locals
        self.env.update_locals(env_update)


        # run the prelude code
        if prelude:
            self.env.run_script(prelude)
            prompt_chunks.append(f'The following prelude code was run in the environment:\n```{prelude}```')


        #based on the locals+prelude, update the docstring for the run method to include descriptions about what is available in the environment
        #TODO: this is illegal...
        # self.run.__doc__ += f'\n\nThe following variables are available in the environment:\n\n' + '\n'.join(prompt_chunks)

        # if locals and prelude were empty, the prompt should say this is a fresh instance and anything needs to be imported

    @tool()
    def run(self, code:str) -> str:
        """
        Runs python code in a python environment.

        The environment is persistent between runs, so any variables created will be available in subsequent runs.
        The only visible effects of this tool are from output to stdout/stderr. If you want to view a result, you MUST print it.

        Args:
            code (str): The code to run

        Returns:
            str: The stdout output of the code
        """
        out, err = self.env.run_script(code)
        if err:
            raise Exception(err)

        return out
    

# @tool()
# def _man(objname:str) -> str:
#     ...
import pydoc
def pyman(obj:Any) -> str:
    """
    Get the documentation for an object in a python environment
    """
    return pydoc.render_doc(obj, renderer=pydoc.plaintext)